#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class TentacleRequirementsTree:
    """
    Tree like structure used to keep track of tentacles nested requirements
    """
    def __init__(self, tentacle, config_name, tentacle_class=None):
        self.tentacle = tentacle
        self.tentacle_class = tentacle_class or self.tentacle.__class__
        self.config_name = config_name
        self.tentacle_config = None     # set when necessary by synchronize_tentacles_config()
        self.nested_requirements = []

    def get_requirement(self, tentacle_class, config_name):
        for requirement in self.nested_requirements:
            if requirement.tentacle_class is tentacle_class and requirement.config_name == config_name:
                return requirement
        return None

    def add_requirement(self, requirement):
        """
        Adds a requirement to nested_requirements
        :param requirement: the requirement to add
        :return: True if nested_requirements did not already contain requirement
        """
        if self.get_requirement(requirement.tentacle_class, requirement.config_name) is None:
            self.nested_requirements.append(requirement)
            return True
        return False

    def get_all_required_tentacles(self, include_self):
        return [req.tentacle for req in self.get_all_nested_requirements(include_self)]

    def synchronize_tentacles_config(self):
        self.tentacle_config = self.tentacle.specific_config if hasattr(self.tentacle, "specific_config") \
            else self.tentacle.trading_config
        for nested_requirement in self.nested_requirements:
            nested_requirement.synchronize_tentacles_config()

    def get_all_nested_requirements(self, include_self):
        if self.tentacle is None:
            raise RuntimeError(f"Impossible to call get_all_nested_requirements() when the tentacle attribute is None")
        requirements = [self] if include_self else []
        for nested_requirement in self.nested_requirements:
            requirements += nested_requirement.get_all_nested_requirements(True)
        return requirements

    def includes_nested_requirements(self, other):
        for req in other.nested_requirements:
            if req not in self.nested_requirements:
                return False
        return True

    def summary(self):
        """
        Returns a clone of this tentacle requirement with your references to the tentacle instance
        """
        summary_requirement = TentacleRequirementsTree(self.tentacle, self.config_name, tentacle_class=self.tentacle_class)
        summary_requirement.tentacle = None
        summary_requirement.tentacle_config = self.tentacle_config
        summary_requirement.nested_requirements = [req.summary() for req in self.nested_requirements]
        return summary_requirement

    def __eq__(self, other):
        if other is None or \
           self.tentacle_class is not other.tentacle_class or \
           self.config_name != other.config_name:
            return False
        # ensure every requirement in self is contained in other's
        return other.includes_nested_requirements(self)

    def __str__(self):
        return f"{self.__class__.__name__} with tentacle: {self.tentacle_class.__name__}, " \
               f"config_name: {self.config_name}, " \
               f"{len(self.nested_requirements)} nested_requirements: " \
               f"[{[str(req) for req in self.nested_requirements]}]"



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
from shutil import rmtree

import aiohttp
import pytest
from os import path

import octobot_tentacles_manager.loaders as loaders
import octobot_tentacles_manager.models as models
import octobot_tentacles_manager.util as util
import octobot_tentacles_manager.api as api
from octobot_tentacles_manager.configuration import TentaclesSetupConfiguration
import octobot_tentacles_manager.constants as constants

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test__set_activation_using_default_config():
    _cleanup()
    async with aiohttp.ClientSession() as session:
        await api.install_all_tentacles(_tentacles_local_path(), aiohttp_session=session)
    tentacle_setup_config = TentaclesSetupConfiguration()
    tentacles = list(loaders.get_tentacle_classes().values())
    tentacle_setup_config._update_tentacles_setup_config(tentacles)

    tentacle_root_path = tentacles[0].tentacle_root_path
    tentacle_types = sorted(util.tentacle_explorer._get_tentacle_types(tentacle_root_path), key=lambda x: x.__str__())

    new_tentacles = []
    # Add Evaluators
    # TA one
    testTA = models.Tentacle(tentacles[0].tentacle_root_path, "testTA",
                             models.TentacleType(os.path.join(constants.TENTACLES_EVALUATOR_PATH,
                                                              constants.TENTACLES_EVALUATOR_TA_PATH)))
    testTA.tentacle_class_names = ["testTA"]
    new_tentacles.append(testTA)
    # Social one
    testSocial = models.Tentacle(tentacles[0].tentacle_root_path, "testSocial", tentacle_types[2])
    testSocial.tentacle_class_names = ["testSocial"]
    new_tentacles.append(testSocial)
    # RealTime one
    testRT = models.Tentacle(tentacles[0].tentacle_root_path, "testRT", tentacle_types[1])
    testRT.tentacle_class_names = ["testRT"]
    new_tentacles.append(testRT)
    # Strategy one
    testStrategy = models.Tentacle(tentacles[0].tentacle_root_path, "testStrategy", tentacle_types[3])
    testStrategy.tentacle_class_names = ["testStrategy"]
    new_tentacles.append(testStrategy)
    # Util one
    testUtil = models.Tentacle(tentacles[0].tentacle_root_path, "testUtil", tentacle_types[4])
    testUtil.tentacle_class_names = ["testUtil"]
    new_tentacles.append(testUtil)

    # Add Trading mode
    # Exchange
    testExchange = models.Tentacle(tentacles[0].tentacle_root_path, "testExchange",
                                   models.TentacleType(os.path.join(constants.TENTACLES_TRADING_PATH,
                                                                    constants.TENTACLES_TRADING_EXCHANGE_PATH)))
    testExchange.tentacle_class_names = ["testExchange"]
    new_tentacles.append(testExchange)
    # Mode
    testMode = models.Tentacle(tentacles[0].tentacle_root_path, "testMode", tentacle_types[8])
    testMode.tentacle_class_names = ["testMode"]
    new_tentacles.append(testMode)

    # Add backtesting
    testBacktesting = models.Tentacle(tentacles[0].tentacle_root_path, "testBacktesting", tentacle_types[0])
    testBacktesting.tentacle_class_names = ["testBacktesting"]
    new_tentacles.append(testBacktesting)

    # Add Meta
    testMetaKeyword = models.Tentacle(tentacles[0].tentacle_root_path, "testMetaKeyword", tentacle_types[5])
    testMetaKeyword.tentacle_class_names = ["testMetaKeyword"]
    new_tentacles.append(testMetaKeyword)

    # Add Service
    testServiceBase = models.Tentacle(tentacles[0].tentacle_root_path, "testServiceBase", tentacle_types[6])
    testServiceBase.tentacle_class_names = ["testServiceBase"]
    testServiceFeed = models.Tentacle(tentacles[0].tentacle_root_path, "testServiceFeed", tentacle_types[7])
    testServiceFeed.tentacle_class_names = ["testServiceFeed"]
    new_tentacles.append(testServiceFeed)
    new_tentacles.append(testServiceBase)

    tentacles += new_tentacles
    tentacle_setup_config._update_tentacles_setup_config(tentacles, newly_installed_tentacles=new_tentacles)

    # Evaluator are disabled by default except for Util
    assert not tentacle_setup_config.tentacles_activation[constants.TENTACLES_EVALUATOR_PATH]["testTA"]
    assert not tentacle_setup_config.tentacles_activation[constants.TENTACLES_EVALUATOR_PATH]["testSocial"]
    assert not tentacle_setup_config.tentacles_activation[constants.TENTACLES_EVALUATOR_PATH]["testStrategy"]
    assert tentacle_setup_config.tentacles_activation[constants.TENTACLES_EVALUATOR_PATH]["testUtil"]

    # Trading
    assert tentacle_setup_config.tentacles_activation[constants.TENTACLES_TRADING_PATH]["testExchange"]
    assert not tentacle_setup_config.tentacles_activation[constants.TENTACLES_TRADING_PATH]["testMode"]

    # Backtesting
    assert tentacle_setup_config.tentacles_activation[constants.TENTACLES_BACKTESTING_PATH]["testBacktesting"]

    # Meta
    assert tentacle_setup_config.tentacles_activation[constants.TENTACLES_META_PATH]["testMetaKeyword"]

    # Services
    assert tentacle_setup_config.tentacles_activation[constants.TENTACLES_SERVICES_PATH]["testServiceBase"]
    assert tentacle_setup_config.tentacles_activation[constants.TENTACLES_SERVICES_PATH]["testServiceFeed"]

    _cleanup()


def _tentacles_local_path():
    return path.join("tests", "static", "tentacles.zip")


def _cleanup():
    if path.exists(constants.TENTACLES_PATH):
        rmtree(constants.TENTACLES_PATH)
    if path.exists(constants.TENTACLE_CONFIG_FILE_NAME):
        os.remove(constants.TENTACLE_CONFIG_FILE_NAME)
    if path.exists(constants.USER_REFERENCE_TENTACLE_CONFIG_PATH):
        rmtree(constants.USER_REFERENCE_TENTACLE_CONFIG_PATH)
